@@ -24,6 +24,7 @@ import (
 	"reflect"
 	"runtime"
 	"runtime/debug"
+	"slices"
 	"sort"
 	"strings"
 	"sync"
@@ -36,14 +37,16 @@ import (
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/filecache"
+	"golang.org/x/tools/gopls/internal/label"
 	"golang.org/x/tools/gopls/internal/progress"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/frob"
+	"golang.org/x/tools/gopls/internal/util/moremaps"
+	"golang.org/x/tools/internal/analysisinternal"
 	"golang.org/x/tools/internal/event"
-	"golang.org/x/tools/internal/event/tag"
 	"golang.org/x/tools/internal/facts"
 	"golang.org/x/tools/internal/gcimporter"
 	"golang.org/x/tools/internal/typesinternal"
@@ -178,8 +181,6 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 
 	var tagStr string // sorted comma-separated list of PackageIDs
 	{
-		// TODO(adonovan): replace with a generic map[S]any -> string
-		// function in the tag package, and use maps.Keys + slices.Sort.
 		keys := make([]string, 0, len(pkgs))
 		for id := range pkgs {
 			keys = append(keys, string(id))
@@ -187,36 +188,36 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 		sort.Strings(keys)
 		tagStr = strings.Join(keys, ",")
 	}
-	ctx, done := event.Start(ctx, "snapshot.Analyze", tag.Package.Of(tagStr))
+	ctx, done := event.Start(ctx, "snapshot.Analyze", label.Package.Of(tagStr))
 	defer done()
 
 	// Filter and sort enabled root analyzers.
 	// A disabled analyzer may still be run if required by another.
 	toSrc := make(map[*analysis.Analyzer]*settings.Analyzer)
-	var enabled []*analysis.Analyzer // enabled subset + transitive requirements
+	var enabledAnalyzers []*analysis.Analyzer // enabled subset + transitive requirements
 	for _, a := range analyzers {
-		if a.IsEnabled(s.Options()) {
-			toSrc[a.Analyzer] = a
-			enabled = append(enabled, a.Analyzer)
+		if enabled, ok := s.Options().Analyses[a.Analyzer().Name]; enabled || !ok && a.EnabledByDefault() {
+			toSrc[a.Analyzer()] = a
+			enabledAnalyzers = append(enabledAnalyzers, a.Analyzer())
 		}
 	}
-	sort.Slice(enabled, func(i, j int) bool {
-		return enabled[i].Name < enabled[j].Name
+	sort.Slice(enabledAnalyzers, func(i, j int) bool {
+		return enabledAnalyzers[i].Name < enabledAnalyzers[j].Name
 	})
 	analyzers = nil // prevent accidental use
 
-	enabled = requiredAnalyzers(enabled)
+	enabledAnalyzers = requiredAnalyzers(enabledAnalyzers)
 
 	// Perform basic sanity checks.
 	// (Ideally we would do this only once.)
-	if err := analysis.Validate(enabled); err != nil {
+	if err := analysis.Validate(enabledAnalyzers); err != nil {
 		return nil, fmt.Errorf("invalid analyzer configuration: %v", err)
 	}
 
 	stableNames := make(map[*analysis.Analyzer]string)
 
 	var facty []*analysis.Analyzer // facty subset of enabled + transitive requirements
-	for _, a := range enabled {
+	for _, a := range enabledAnalyzers {
 		// TODO(adonovan): reject duplicate stable names (very unlikely).
 		stableNames[a] = stableName(a)
 
@@ -255,6 +256,8 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 
 			an = &analysisNode{
 				fset:        fset,
+				fsource:     struct{ file.Source }{s}, // expose only ReadFile
+				viewType:    s.View().Type(),
 				mp:          mp,
 				analyzers:   facty, // all nodes run at least the facty analyzers
 				allDeps:     make(map[PackagePath]*analysisNode),
@@ -304,10 +307,10 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 		}
 		// Add edge from predecessor.
 		if from != nil {
-			atomic.AddInt32(&from.unfinishedSuccs, 1) // TODO(adonovan): use generics
+			from.unfinishedSuccs.Add(+1) // incref
 			an.preds = append(an.preds, from)
 		}
-		atomic.AddInt32(&an.unfinishedPreds, 1)
+		an.unfinishedPreds.Add(+1)
 		return an, nil
 	}
 
@@ -318,7 +321,7 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 		if err != nil {
 			return nil, err
 		}
-		root.analyzers = enabled
+		root.analyzers = enabledAnalyzers
 		roots = append(roots, root)
 	}
 
@@ -388,7 +391,7 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 	// prevents workers from enqeuing, and thus finishing, and thus allowing the
 	// group to make progress: deadlock.
 	limiter := make(chan unit, runtime.GOMAXPROCS(0))
-	var completed int64
+	var completed atomic.Int64
 
 	var enqueue func(*analysisNode)
 	enqueue = func(an *analysisNode) {
@@ -400,13 +403,13 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 			if err != nil {
 				return err // cancelled, or failed to produce a package
 			}
-			maybeReport(atomic.AddInt64(&completed, 1))
+			maybeReport(completed.Add(1))
 			an.summary = summary
 
 			// Notify each waiting predecessor,
 			// and enqueue it when it becomes a leaf.
 			for _, pred := range an.preds {
-				if atomic.AddInt32(&pred.unfinishedSuccs, -1) == 0 {
+				if pred.unfinishedSuccs.Add(-1) == 0 { // decref
 					enqueue(pred)
 				}
 			}
@@ -428,6 +431,18 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 		return nil, err // cancelled, or failed to produce a package
 	}
 
+	// Inv: all root nodes now have a summary (#66732).
+	//
+	// We know this is falsified empirically. This means either
+	// the summary was "successfully" set to nil (above), or there
+	// is a problem with the graph such the enqueuing leaves does
+	// not lead to completion of roots (or an error).
+	for _, root := range roots {
+		if root.summary == nil {
+			bug.Report("root analysisNode has nil summary")
+		}
+	}
+
 	// Report diagnostics only from enabled actions that succeeded.
 	// Errors from creating or analyzing packages are ignored.
 	// Diagnostics are reported in the order of the analyzers argument.
@@ -440,7 +455,7 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 	// results, we should propagate the per-action errors.
 	var results []*Diagnostic
 	for _, root := range roots {
-		for _, a := range enabled {
+		for _, a := range enabledAnalyzers {
 			// Skip analyzers that were added only to
 			// fulfil requirements of the original set.
 			srcAnalyzer, ok := toSrc[a]
@@ -459,6 +474,7 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 			}
 
 			// Inv: root.summary is the successful result of run (via runCached).
+			// TODO(adonovan): fix: root.summary is sometimes nil! (#66732).
 			summary, ok := root.summary.Actions[stableNames[a]]
 			if summary == nil {
 				panic(fmt.Sprintf("analyzeSummary.Actions[%q] = (nil, %t); got %v (#60551)",
@@ -476,7 +492,7 @@ func (s *Snapshot) Analyze(ctx context.Context, pkgs map[PackageID]*metadata.Pac
 }
 
 func (an *analysisNode) decrefPreds() {
-	if atomic.AddInt32(&an.unfinishedPreds, -1) == 0 {
+	if an.unfinishedPreds.Add(-1) == 0 {
 		an.summary.Actions = nil
 	}
 }
@@ -506,13 +522,15 @@ func (an *analysisNode) decrefPreds() {
 // type-checking and analyzing syntax (miss).
 type analysisNode struct {
 	fset            *token.FileSet              // file set shared by entire batch (DAG)
+	fsource         file.Source                 // Snapshot.ReadFile, for use by Pass.ReadFile
+	viewType        ViewType                    // type of view
 	mp              *metadata.Package           // metadata for this package
 	files           []file.Handle               // contents of CompiledGoFiles
 	analyzers       []*analysis.Analyzer        // set of analyzers to run
 	preds           []*analysisNode             // graph edges:
 	succs           map[PackageID]*analysisNode //   (preds -> self -> succs)
-	unfinishedSuccs int32
-	unfinishedPreds int32                         // effectively a summary.Actions refcount
+	unfinishedSuccs atomic.Int32
+	unfinishedPreds atomic.Int32                  // effectively a summary.Actions refcount
 	allDeps         map[PackagePath]*analysisNode // all dependencies including self
 	exportDeps      map[PackagePath]*analysisNode // subset of allDeps ref'd by export data (+self)
 	summary         *analyzeSummary               // serializable result of analyzing this package
@@ -521,6 +539,9 @@ type analysisNode struct {
 	typesOnce sync.Once      // guards lazy population of types and typesErr fields
 	types     *types.Package // type information lazily imported from summary
 	typesErr  error          // an error producing type information
+
+	depHashOnce sync.Once
+	_depHash    file.Hash // memoized hash of data affecting dependents
 }
 
 func (an *analysisNode) String() string { return string(an.mp.ID) }
@@ -579,6 +600,40 @@ func (an *analysisNode) _import() (*types.Package, error) {
 	return an.types, an.typesErr
 }
 
+// depHash computes the hash of node information that may affect other nodes
+// depending on this node: the package path, export hash, and action results.
+//
+// The result is memoized to avoid redundant work when analysing multiple
+// dependents.
+func (an *analysisNode) depHash() file.Hash {
+	an.depHashOnce.Do(func() {
+		hasher := sha256.New()
+		fmt.Fprintf(hasher, "dep: %s\n", an.mp.PkgPath)
+		fmt.Fprintf(hasher, "export: %s\n", an.summary.DeepExportHash)
+
+		// action results: errors and facts
+		actions := an.summary.Actions
+		names := make([]string, 0, len(actions))
+		for name := range actions {
+			names = append(names, name)
+		}
+		sort.Strings(names)
+		for _, name := range names {
+			summary := actions[name]
+			fmt.Fprintf(hasher, "action %s\n", name)
+			if summary.Err != "" {
+				fmt.Fprintf(hasher, "error %s\n", summary.Err)
+			} else {
+				fmt.Fprintf(hasher, "facts %s\n", summary.FactsHash)
+				// We can safely omit summary.diagnostics
+				// from the key since they have no downstream effect.
+			}
+		}
+		hasher.Sum(an._depHash[:0])
+	})
+	return an._depHash
+}
+
 // analyzeSummary is a gob-serializable summary of successfully
 // applying a list of analyzers to a package.
 type analyzeSummary struct {
@@ -665,6 +720,9 @@ func (an *analysisNode) runCached(ctx context.Context) (*analyzeSummary, error)
 	if data, err := filecache.Get(cacheKind, key); err == nil {
 		// cache hit
 		analyzeSummaryCodec.Decode(data, &summary)
+		if summary == nil { // debugging #66732
+			bug.Reportf("analyzeSummaryCodec.Decode yielded nil *analyzeSummary")
+		}
 	} else if err != filecache.ErrNotFound {
 		return nil, bug.Errorf("internal error reading shared cache: %v", err)
 	} else {
@@ -674,8 +732,11 @@ func (an *analysisNode) runCached(ctx context.Context) (*analyzeSummary, error)
 		if err != nil {
 			return nil, err
 		}
+		if summary == nil { // debugging #66732 (can't happen)
+			bug.Reportf("analyzeNode.run returned nil *analyzeSummary")
+		}
 
-		atomic.AddInt32(&an.unfinishedPreds, +1) // incref
+		an.unfinishedPreds.Add(+1) // incref
 		go func() {
 			defer an.decrefPreds() //decref
 
@@ -720,6 +781,8 @@ func (an *analysisNode) cacheKey() [sha256.Size]byte {
 	// package metadata
 	mp := an.mp
 	fmt.Fprintf(hasher, "package: %s %s %s\n", mp.ID, mp.Name, mp.PkgPath)
+	fmt.Fprintf(hasher, "viewtype: %s\n", an.viewType) // (affects diagnostics)
+
 	// We can ignore m.DepsBy{Pkg,Import}Path: although the logic
 	// uses those fields, we account for them by hashing vdeps.
 
@@ -743,34 +806,9 @@ func (an *analysisNode) cacheKey() [sha256.Size]byte {
 	}
 
 	// vdeps, in PackageID order
-	depIDs := make([]string, 0, len(an.succs))
-	for depID := range an.succs {
-		depIDs = append(depIDs, string(depID))
-	}
-	sort.Strings(depIDs) // TODO(adonovan): avoid conversions by using slices.Sort[PackageID]
-	for _, depID := range depIDs {
-		vdep := an.succs[PackageID(depID)]
-		fmt.Fprintf(hasher, "dep: %s\n", vdep.mp.PkgPath)
-		fmt.Fprintf(hasher, "export: %s\n", vdep.summary.DeepExportHash)
-
-		// action results: errors and facts
-		actions := vdep.summary.Actions
-		names := make([]string, 0, len(actions))
-		for name := range actions {
-			names = append(names, name)
-		}
-		sort.Strings(names)
-		for _, name := range names {
-			summary := actions[name]
-			fmt.Fprintf(hasher, "action %s\n", name)
-			if summary.Err != "" {
-				fmt.Fprintf(hasher, "error %s\n", summary.Err)
-			} else {
-				fmt.Fprintf(hasher, "facts %s\n", summary.FactsHash)
-				// We can safely omit summary.diagnostics
-				// from the key since they have no downstream effect.
-			}
-		}
+	for _, vdep := range moremaps.Sorted(an.succs) {
+		hash := vdep.depHash()
+		hasher.Write(hash[:])
 	}
 
 	var hash [sha256.Size]byte
@@ -868,6 +906,7 @@ func (an *analysisNode) run(ctx context.Context) (*analyzeSummary, error) {
 			}
 			act = &action{
 				a:          a,
+				fsource:    an.fsource,
 				stableName: an.stableNames[a],
 				pkg:        pkg,
 				vdeps:      an.succs,
@@ -885,7 +924,7 @@ func (an *analysisNode) run(ctx context.Context) (*analyzeSummary, error) {
 	}
 
 	// Execute the graph in parallel.
-	execActions(roots)
+	execActions(ctx, roots)
 	// Inv: each root's summary is set (whether success or error).
 
 	// Don't return (or cache) the result in case of cancellation.
@@ -1002,7 +1041,7 @@ func (an *analysisNode) typeCheck(parsed []*parsego.File) *analysisPackage {
 			}
 
 			// (Duplicates logic from check.go.)
-			if !metadata.IsValidImport(an.mp.PkgPath, dep.mp.PkgPath) {
+			if !metadata.IsValidImport(an.mp.PkgPath, dep.mp.PkgPath, an.viewType != GoPackagesDriverView) {
 				return nil, fmt.Errorf("invalid use of internal package %s", importPath)
 			}
 
@@ -1013,11 +1052,8 @@ func (an *analysisNode) typeCheck(parsed []*parsego.File) *analysisPackage {
 	// Set Go dialect.
 	if mp.Module != nil && mp.Module.GoVersion != "" {
 		goVersion := "go" + mp.Module.GoVersion
-		// types.NewChecker panics if GoVersion is invalid.
-		// An unparsable mod file should probably stop us
-		// before we get here, but double check just in case.
-		if goVersionRx.MatchString(goVersion) {
-			typesinternal.SetGoVersion(cfg, goVersion)
+		if validGoVersion(goVersion) {
+			cfg.GoVersion = goVersion
 		}
 	}
 
@@ -1121,7 +1157,8 @@ type analysisPackage struct {
 type action struct {
 	once       sync.Once
 	a          *analysis.Analyzer
-	stableName string // cross-process stable name of analyzer
+	fsource    file.Source // Snapshot.ReadFile, for Pass.ReadFile
+	stableName string      // cross-process stable name of analyzer
 	pkg        *analysisPackage
 	hdeps      []*action                   // horizontal dependencies
 	vdeps      map[PackageID]*analysisNode // vertical dependencies
@@ -1138,7 +1175,7 @@ func (act *action) String() string {
 
 // execActions executes a set of action graph nodes in parallel.
 // Postcondition: each action.summary is set, even in case of error.
-func execActions(actions []*action) {
+func execActions(ctx context.Context, actions []*action) {
 	var wg sync.WaitGroup
 	for _, act := range actions {
 		act := act
@@ -1146,8 +1183,8 @@ func execActions(actions []*action) {
 		go func() {
 			defer wg.Done()
 			act.once.Do(func() {
-				execActions(act.hdeps) // analyze "horizontal" dependencies
-				act.result, act.summary, act.err = act.exec()
+				execActions(ctx, act.hdeps) // analyze "horizontal" dependencies
+				act.result, act.summary, act.err = act.exec(ctx)
 				if act.err != nil {
 					act.summary = &actionSummary{Err: act.err.Error()}
 					// TODO(adonovan): suppress logging. But
@@ -1171,7 +1208,7 @@ func execActions(actions []*action) {
 // along with its (serializable) facts and diagnostics.
 // Or it returns an error if the analyzer did not run to
 // completion and deliver a valid result.
-func (act *action) exec() (interface{}, *actionSummary, error) {
+func (act *action) exec(ctx context.Context) (any, *actionSummary, error) {
 	analyzer := act.a
 	pkg := act.pkg
 
@@ -1270,53 +1307,114 @@ func (act *action) exec() (interface{}, *actionSummary, error) {
 		factFilter[reflect.TypeOf(f)] = true
 	}
 
-	// If the package contains "fixed" files, it's not necessarily an error if we
-	// can't convert positions.
-	hasFixedFiles := false
-	for _, p := range pkg.parsed {
-		if p.Fixed() {
-			hasFixedFiles = true
-			break
-		}
-	}
-
 	// posToLocation converts from token.Pos to protocol form.
-	// TODO(adonovan): improve error messages.
 	posToLocation := func(start, end token.Pos) (protocol.Location, error) {
 		tokFile := pkg.fset.File(start)
 
+		// Find existing mapper by file name.
+		// (Don't require an exact token.File match
+		// as the analyzer may have re-parsed the file.)
+		var (
+			mapper *protocol.Mapper
+			fixed  bool
+		)
 		for _, p := range pkg.parsed {
-			if p.Tok == tokFile {
-				if end == token.NoPos {
-					end = start
-				}
-				return p.PosLocation(start, end)
+			if p.Tok.Name() == tokFile.Name() {
+				mapper = p.Mapper
+				fixed = p.Fixed() // suppress some assertions after parser recovery
+				break
 			}
 		}
-		errorf := bug.Errorf
-		if hasFixedFiles {
-			errorf = fmt.Errorf
+		if mapper == nil {
+			// The start position was not among the package's parsed
+			// Go files, indicating that the analyzer added new files
+			// to the FileSet.
+			//
+			// For example, the cgocall analyzer re-parses and
+			// type-checks some of the files in a special environment;
+			// and asmdecl and other low-level runtime analyzers call
+			// ReadFile to parse non-Go files.
+			// (This is a supported feature, documented at go/analysis.)
+			//
+			// In principle these files could be:
+			//
+			// - OtherFiles (non-Go files such as asm).
+			//   However, we set Pass.OtherFiles=[] because
+			//   gopls won't service "diagnose" requests
+			//   for non-Go files, so there's no point
+			//   reporting diagnostics in them.
+			//
+			// - IgnoredFiles (files tagged for other configs).
+			//   However, we set Pass.IgnoredFiles=[] because,
+			//   in most cases, zero-config gopls should create
+			//   another view that covers these files.
+			//
+			// - Referents of //line directives, as in cgo packages.
+			//   The file names in this case are not known a priori.
+			//   gopls generally tries to avoid honoring line directives,
+			//   but analyzers such as cgocall may honor them.
+			//
+			// In short, it's unclear how this can be reached
+			// other than due to an analyzer bug.
+			return protocol.Location{}, bug.Errorf("diagnostic location is not among files of package: %s", tokFile.Name())
 		}
-		return protocol.Location{}, errorf("token.Pos not within package")
+		// Inv: mapper != nil
+
+		if end == token.NoPos {
+			end = start
+		}
+
+		// debugging #64547
+		fileStart := token.Pos(tokFile.Base())
+		fileEnd := fileStart + token.Pos(tokFile.Size())
+		if start < fileStart {
+			if !fixed {
+				bug.Reportf("start < start of file")
+			}
+			start = fileStart
+		}
+		if end < start {
+			// This can happen if End is zero (#66683)
+			// or a small positive displacement from zero
+			// due to recursive Node.End() computation.
+			// This usually arises from poor parser recovery
+			// of an incomplete term at EOF.
+			if !fixed {
+				bug.Reportf("end < start of file")
+			}
+			end = fileEnd
+		}
+		if end > fileEnd+1 {
+			if !fixed {
+				bug.Reportf("end > end of file + 1")
+			}
+			end = fileEnd
+		}
+
+		return mapper.PosLocation(tokFile, start, end)
 	}
 
 	// Now run the (pkg, analyzer) action.
 	var diagnostics []gobDiagnostic
+
 	pass := &analysis.Pass{
-		Analyzer:   analyzer,
-		Fset:       pkg.fset,
-		Files:      pkg.files,
-		Pkg:        pkg.types,
-		TypesInfo:  pkg.typesInfo,
-		TypesSizes: pkg.typesSizes,
-		TypeErrors: pkg.typeErrors,
-		ResultOf:   inputs,
+		Analyzer:     analyzer,
+		Fset:         pkg.fset,
+		Files:        pkg.files,
+		OtherFiles:   nil, // since gopls doesn't handle non-Go (e.g. asm) files
+		IgnoredFiles: nil, // zero-config gopls should analyze these files in another view
+		Pkg:          pkg.types,
+		TypesInfo:    pkg.typesInfo,
+		TypesSizes:   pkg.typesSizes,
+		TypeErrors:   pkg.typeErrors,
+		ResultOf:     inputs,
 		Report: func(d analysis.Diagnostic) {
 			diagnostic, err := toGobDiagnostic(posToLocation, analyzer, d)
 			if err != nil {
-				if !hasFixedFiles {
-					bug.Reportf("internal error converting diagnostic from analyzer %q: %v", analyzer.Name, err)
-				}
+				// Don't bug.Report here: these errors all originate in
+				// posToLocation, and we can more accurately discriminate
+				// severe errors from benign ones in that function.
+				event.Error(ctx, fmt.Sprintf("internal error converting diagnostic from analyzer %q", analyzer.Name), err)
 				return
 			}
 			diagnostics = append(diagnostics, diagnostic)
@@ -1329,6 +1427,29 @@ func (act *action) exec() (interface{}, *actionSummary, error) {
 		AllPackageFacts:   func() []analysis.PackageFact { return factset.AllPackageFacts(factFilter) },
 	}
 
+	pass.ReadFile = func(filename string) ([]byte, error) {
+		// Read file from snapshot, to ensure reads are consistent.
+		//
+		// TODO(adonovan): make the dependency analysis sound by
+		// incorporating these additional files into the the analysis
+		// hash. This requires either (a) preemptively reading and
+		// hashing a potentially large number of mostly irrelevant
+		// files; or (b) some kind of dynamic dependency discovery
+		// system like used in Bazel for C++ headers. Neither entices.
+		if err := analysisinternal.CheckReadable(pass, filename); err != nil {
+			return nil, err
+		}
+		h, err := act.fsource.ReadFile(ctx, protocol.URIFromPath(filename))
+		if err != nil {
+			return nil, err
+		}
+		content, err := h.Content()
+		if err != nil {
+			return nil, err // file doesn't exist
+		}
+		return slices.Clone(content), nil // follow ownership of os.ReadFile
+	}
+
 	// Recover from panics (only) within the analyzer logic.
 	// (Use an anonymous function to limit the recover scope.)
 	var result interface{}
@@ -1402,7 +1523,7 @@ type LabelDuration struct {
 	Duration time.Duration
 }
 
-// AnalyzerTimes returns the accumulated time spent in each Analyzer's
+// AnalyzerRunTimes returns the accumulated time spent in each Analyzer's
 // Run function since process start, in descending order.
 func AnalyzerRunTimes() []LabelDuration {
 	analyzerRunTimesMu.Lock()
