@@ -10,6 +10,7 @@ import (
 	"errors"
 	"fmt"
 	"path/filepath"
+	"slices"
 	"sort"
 	"strings"
 	"sync/atomic"
@@ -23,7 +24,6 @@ import (
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/immutable"
 	"golang.org/x/tools/gopls/internal/util/pathutil"
-	"golang.org/x/tools/gopls/internal/util/slices"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/gocommand"
 	"golang.org/x/tools/internal/packagesinternal"
@@ -118,16 +118,13 @@ func (s *Snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 
 	startTime := time.Now()
 
-	flags := LoadWorkspace
-	if allowNetwork {
-		flags |= AllowNetwork
-	}
-	_, inv, cleanup, err := s.goCommandInvocation(ctx, flags, &gocommand.Invocation{
+	inv, cleanupInvocation, err := s.GoCommandInvocation(allowNetwork, &gocommand.Invocation{
 		WorkingDir: s.view.root.Path(),
 	})
 	if err != nil {
 		return err
 	}
+	defer cleanupInvocation()
 
 	// Set a last resort deadline on packages.Load since it calls the go
 	// command, which may hang indefinitely if it has a bug. golang/go#42132
@@ -137,7 +134,6 @@ func (s *Snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 
 	cfg := s.config(ctx, inv)
 	pkgs, err := packages.Load(cfg, query...)
-	cleanup()
 
 	// If the context was canceled, return early. Otherwise, we might be
 	// type-checking an incomplete result. Check the context directly,
@@ -266,7 +262,7 @@ func (s *Snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 		if allFilesExcluded(pkg.GoFiles, filterFunc) {
 			continue
 		}
-		buildMetadata(newMetadata, pkg, cfg.Dir, standalone)
+		buildMetadata(newMetadata, pkg, cfg.Dir, standalone, s.view.typ != GoPackagesDriverView)
 	}
 
 	s.mu.Lock()
@@ -358,16 +354,11 @@ func (m *moduleErrorMap) Error() string {
 // Returns the metadata.Package that was built (or which was already present in
 // updates), or nil if the package could not be built. Notably, the resulting
 // metadata.Package may have an ID that differs from pkg.ID.
-func buildMetadata(updates map[PackageID]*metadata.Package, pkg *packages.Package, loadDir string, standalone bool) *metadata.Package {
+func buildMetadata(updates map[PackageID]*metadata.Package, pkg *packages.Package, loadDir string, standalone, goListView bool) *metadata.Package {
 	// Allow for multiple ad-hoc packages in the workspace (see #47584).
 	pkgPath := PackagePath(pkg.PkgPath)
 	id := PackageID(pkg.ID)
 
-	// debugging #60890
-	if pkg.PkgPath == "unsafe" && pkg.ID != "unsafe" {
-		bug.Reportf("PackagePath \"unsafe\" with ID %q", pkg.ID)
-	}
-
 	if metadata.IsCommandLineArguments(id) {
 		var f string // file to use as disambiguating suffix
 		if len(pkg.CompiledGoFiles) > 0 {
@@ -419,25 +410,17 @@ func buildMetadata(updates map[PackageID]*metadata.Package, pkg *packages.Packag
 		Standalone: standalone,
 	}
 
-	// debugging #60890
-	if mp.PkgPath == "unsafe" && mp.ID != "unsafe" {
-		bug.Reportf("PackagePath \"unsafe\" with ID %q", mp.ID)
-	}
-
 	updates[id] = mp
 
-	for _, filename := range pkg.CompiledGoFiles {
-		uri := protocol.URIFromPath(filename)
-		mp.CompiledGoFiles = append(mp.CompiledGoFiles, uri)
-	}
-	for _, filename := range pkg.GoFiles {
-		uri := protocol.URIFromPath(filename)
-		mp.GoFiles = append(mp.GoFiles, uri)
-	}
-	for _, filename := range pkg.IgnoredFiles {
-		uri := protocol.URIFromPath(filename)
-		mp.IgnoredFiles = append(mp.IgnoredFiles, uri)
+	copyURIs := func(dst *[]protocol.DocumentURI, src []string) {
+		for _, filename := range src {
+			*dst = append(*dst, protocol.URIFromPath(filename))
+		}
 	}
+	copyURIs(&mp.CompiledGoFiles, pkg.CompiledGoFiles)
+	copyURIs(&mp.GoFiles, pkg.GoFiles)
+	copyURIs(&mp.IgnoredFiles, pkg.IgnoredFiles)
+	copyURIs(&mp.OtherFiles, pkg.OtherFiles)
 
 	depsByImpPath := make(map[ImportPath]PackageID)
 	depsByPkgPath := make(map[PackagePath]PackageID)
@@ -524,7 +507,7 @@ func buildMetadata(updates map[PackageID]*metadata.Package, pkg *packages.Packag
 			continue
 		}
 
-		dep := buildMetadata(updates, imported, loadDir, false) // only top level packages can be standalone
+		dep := buildMetadata(updates, imported, loadDir, false, goListView) // only top level packages can be standalone
 
 		// Don't record edges to packages with no name, as they cause trouble for
 		// the importer (golang/go#60952).
@@ -680,7 +663,7 @@ func isWorkspacePackageLocked(ctx context.Context, s *Snapshot, meta *metadata.G
 	//
 	// For module views (of type GoMod or GoWork), packages must in any case be
 	// in a workspace module (enforced below).
-	if !s.view.moduleMode() || !s.Options().ExpandWorkspaceToModule {
+	if !s.view.typ.usesModules() || !s.Options().ExpandWorkspaceToModule {
 		folder := s.view.folder.Dir.Path()
 		inFolder := false
 		for uri := range uris {
@@ -696,7 +679,7 @@ func isWorkspacePackageLocked(ctx context.Context, s *Snapshot, meta *metadata.G
 
 	// In module mode, a workspace package must be contained in a workspace
 	// module.
-	if s.view.moduleMode() {
+	if s.view.typ.usesModules() {
 		var modURI protocol.DocumentURI
 		if pkg.Module != nil {
 			modURI = protocol.URIFromPath(pkg.Module.GoMod)
