@@ -11,12 +11,11 @@ import (
 	"fmt"
 	"log"
 	"os"
-	"strings"
+	"slices"
 
-	"golang.org/x/tools/gopls/internal/lsp/command"
-	"golang.org/x/tools/gopls/internal/lsp/protocol"
+	"golang.org/x/tools/gopls/internal/protocol"
+	"golang.org/x/tools/gopls/internal/protocol/command"
 	"golang.org/x/tools/gopls/internal/server"
-	"golang.org/x/tools/gopls/internal/util/slices"
 	"golang.org/x/tools/internal/tool"
 )
 
@@ -36,11 +35,9 @@ The execute command sends an LSP ExecuteCommand request to gopls,
 with a set of optional JSON argument values.
 Some commands return a result, also JSON.
 
-Available commands are documented at:
-
-	https://github.com/golang/tools/blob/master/gopls/doc/commands.md
-
-This interface is experimental and commands may change or disappear without notice.
+Gopls' command set is defined by the command.Interface type; see
+https://pkg.go.dev/golang.org/x/tools/gopls/internal/protocol/command#Interface.
+It is not a stable interface: commands may change or disappear without notice.
 
 Examples:
 
@@ -58,7 +55,7 @@ func (e *execute) Run(ctx context.Context, args ...string) error {
 		return tool.CommandLineErrorf("execute requires a command name")
 	}
 	cmd := args[0]
-	if !slices.Contains(command.Commands, command.Command(strings.TrimPrefix(cmd, "gopls."))) {
+	if !slices.Contains(command.Commands, command.Command(cmd)) {
 		return tool.CommandLineErrorf("unrecognized command: %s", cmd)
 	}
 
@@ -76,14 +73,13 @@ func (e *execute) Run(ctx context.Context, args ...string) error {
 
 	e.app.editFlags = &e.EditFlags // in case command performs an edit
 
-	cmdDone, onProgress := commandProgress()
-	conn, err := e.app.connect(ctx, onProgress)
+	conn, err := e.app.connect(ctx)
 	if err != nil {
 		return err
 	}
 	defer conn.terminate(ctx)
 
-	res, err := conn.executeCommand(ctx, cmdDone, &protocol.Command{
+	res, err := conn.executeCommand(ctx, &protocol.Command{
 		Command:   cmd,
 		Arguments: jsonArgs,
 	})
@@ -100,55 +96,39 @@ func (e *execute) Run(ctx context.Context, args ...string) error {
 	return nil
 }
 
-// -- shared command helpers --
-
-const cmdProgressToken = "cmd-progress"
-
-// TODO(adonovan): disentangle this from app.connect, and factor with
-// conn.executeCommand used by codelens and execute. Seems like
-// connection needs a way to register and unregister independent
-// handlers, later than at connect time.
-func commandProgress() (<-chan bool, func(p *protocol.ProgressParams)) {
-	cmdDone := make(chan bool, 1)
-	onProgress := func(p *protocol.ProgressParams) {
-		switch v := p.Value.(type) {
+// executeCommand executes a protocol.Command, displaying progress
+// messages and awaiting completion of asynchronous commands.
+func (conn *connection) executeCommand(ctx context.Context, cmd *protocol.Command) (any, error) {
+	endStatus := make(chan string, 1)
+	token, unregister := conn.client.registerProgressHandler(func(params *protocol.ProgressParams) {
+		switch v := params.Value.(type) {
 		case *protocol.WorkDoneProgressReport:
-			// TODO(adonovan): how can we segregate command's stdout and
-			// stderr so that structure is preserved?
-			fmt.Fprintln(os.Stderr, v.Message)
+			fmt.Fprintln(os.Stderr, v.Message) // combined std{out,err}
 
 		case *protocol.WorkDoneProgressEnd:
-			if p.Token == cmdProgressToken {
-				// commandHandler.run sends message = canceled | failed | completed
-				cmdDone <- v.Message == server.CommandCompleted
-			}
+			endStatus <- v.Message // = canceled | failed | completed
 		}
-	}
-	return cmdDone, onProgress
-}
+	})
+	defer unregister()
 
-func (conn *connection) executeCommand(ctx context.Context, done <-chan bool, cmd *protocol.Command) (any, error) {
 	res, err := conn.ExecuteCommand(ctx, &protocol.ExecuteCommandParams{
 		Command:   cmd.Command,
 		Arguments: cmd.Arguments,
 		WorkDoneProgressParams: protocol.WorkDoneProgressParams{
-			WorkDoneToken: cmdProgressToken,
+			WorkDoneToken: token,
 		},
 	})
 	if err != nil {
 		return nil, err
 	}
 
-	// Wait for it to finish (by watching for a progress token).
-	//
-	// In theory this is only necessary for the two async
-	// commands (RunGovulncheck and RunTests), but the tests
-	// fail for Test as well (why?), and there is no cost to
-	// waiting in all cases. TODO(adonovan): investigate.
-	if success := <-done; !success {
-		// TODO(adonovan): suppress this message;
-		// the command's stderr should suffice.
-		return nil, fmt.Errorf("command failed")
+	// Some commands are asynchronous, so clients
+	// must wait for the "end" progress notification.
+	if command.Command(cmd.Command).IsAsync() {
+		status := <-endStatus
+		if status != server.CommandCompleted {
+			return nil, fmt.Errorf("command %s", status)
+		}
 	}
 
 	return res, nil
