@@ -14,7 +14,6 @@ import (
 	"golang.org/x/tools/gopls/internal/golang"
 	"golang.org/x/tools/gopls/internal/golang/completion/snippet"
 	"golang.org/x/tools/gopls/internal/protocol"
-	"golang.org/x/tools/internal/aliases"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/typesinternal"
 )
@@ -51,9 +50,13 @@ func (c *completer) literal(ctx context.Context, literalType types.Type, imp *im
 	//
 	// don't offer "mySlice{}" since we have already added a candidate
 	// of "[]int{}".
-	if is[*types.Named](aliases.Unalias(literalType)) &&
+
+	// TODO(adonovan): think about aliases:
+	// they should probably be treated more like Named.
+	// Should this use Deref not Unpointer?
+	if is[*types.Named](types.Unalias(literalType)) &&
 		expType != nil &&
-		!is[*types.Named](aliases.Unalias(typesinternal.Unpointer(expType))) {
+		!is[*types.Named](types.Unalias(typesinternal.Unpointer(expType))) {
 
 		return
 	}
@@ -196,7 +199,7 @@ func (c *completer) functionLiteral(ctx context.Context, sig *types.Signature, m
 			name = p.Name()
 		)
 
-		if tp, _ := p.Type().(*types.TypeParam); tp != nil && !c.typeParamInScope(tp) {
+		if tp, _ := types.Unalias(p.Type()).(*types.TypeParam); tp != nil && !c.typeParamInScope(tp) {
 			hasTypeParams = true
 		}
 
@@ -287,7 +290,7 @@ func (c *completer) functionLiteral(ctx context.Context, sig *types.Signature, m
 				typeStr = strings.Replace(typeStr, "[]", "...", 1)
 			}
 
-			if tp, _ := p.Type().(*types.TypeParam); tp != nil && !c.typeParamInScope(tp) {
+			if tp, ok := types.Unalias(p.Type()).(*types.TypeParam); ok && !c.typeParamInScope(tp) {
 				snip.WritePlaceholder(func(snip *snippet.Builder) {
 					snip.WriteText(typeStr)
 				})
@@ -308,7 +311,7 @@ func (c *completer) functionLiteral(ctx context.Context, sig *types.Signature, m
 
 	var resultHasTypeParams bool
 	for i := 0; i < results.Len(); i++ {
-		if tp, _ := results.At(i).Type().(*types.TypeParam); tp != nil && !c.typeParamInScope(tp) {
+		if tp, ok := types.Unalias(results.At(i).Type()).(*types.TypeParam); ok && !c.typeParamInScope(tp) {
 			resultHasTypeParams = true
 		}
 	}
@@ -341,7 +344,7 @@ func (c *completer) functionLiteral(ctx context.Context, sig *types.Signature, m
 			}
 			return
 		}
-		if tp, _ := r.Type().(*types.TypeParam); tp != nil && !c.typeParamInScope(tp) {
+		if tp, ok := types.Unalias(r.Type()).(*types.TypeParam); ok && !c.typeParamInScope(tp) {
 			snip.WritePlaceholder(func(snip *snippet.Builder) {
 				snip.WriteText(text)
 			})
@@ -428,11 +431,12 @@ func abbreviateTypeName(s string) string {
 }
 
 // compositeLiteral adds a composite literal completion item for the given typeName.
+// T is an (unnamed, unaliased) struct, array, slice, or map type.
 func (c *completer) compositeLiteral(T types.Type, snip *snippet.Builder, typeName string, matchScore float64, edits []protocol.TextEdit) {
 	snip.WriteText("{")
 	// Don't put the tab stop inside the composite literal curlies "{}"
 	// for structs that have no accessible fields.
-	if strct, ok := T.(*types.Struct); !ok || fieldsAccessible(strct, c.pkg.GetTypes()) {
+	if strct, ok := T.(*types.Struct); !ok || fieldsAccessible(strct, c.pkg.Types()) {
 		snip.WriteFinalTabstop()
 	}
 	snip.WriteText("}")
@@ -512,7 +516,9 @@ func (c *completer) typeNameSnippet(literalType types.Type, qf types.Qualifier)
 	var (
 		snip     snippet.Builder
 		typeName string
-		named, _ = literalType.(*types.Named)
+		// TODO(adonovan): think more about aliases.
+		// They should probably be treated more like Named.
+		named, _ = types.Unalias(literalType).(*types.Named)
 	)
 
 	if named != nil && named.Obj() != nil && named.TypeParams().Len() > 0 && !c.fullyInstantiated(named) {
@@ -559,7 +565,8 @@ func (c *completer) fullyInstantiated(t *types.Named) bool {
 	}
 
 	for i := 0; i < tas.Len(); i++ {
-		switch ta := tas.At(i).(type) {
+		// TODO(adonovan) think about generic aliases.
+		switch ta := types.Unalias(tas.At(i)).(type) {
 		case *types.TypeParam:
 			// A *TypeParam only counts as specified if it is currently in
 			// scope (i.e. we are in a generic definition).
@@ -592,8 +599,3 @@ func (c *completer) typeParamInScope(tp *types.TypeParam) bool {
 	_, foundObj := scope.LookupParent(obj.Name(), c.pos)
 	return obj == foundObj
 }
-
-func is[T any](x any) bool {
-	_, ok := x.(T)
-	return ok
-}
