@@ -11,11 +11,13 @@ import (
 	"errors"
 	"fmt"
 	"io"
+	"log"
 	"os"
 	"path/filepath"
 	"regexp"
 	"runtime"
 	"runtime/pprof"
+	"slices"
 	"sort"
 	"strings"
 	"sync"
@@ -70,6 +72,189 @@ type commandHandler struct {
 	params *protocol.ExecuteCommandParams
 }
 
+func (h *commandHandler) Modules(ctx context.Context, args command.ModulesArgs) (command.ModulesResult, error) {
+	// keepModule filters modules based on the command args
+	keepModule := func(goMod protocol.DocumentURI) bool {
+		// Does the directory enclose the view's go.mod file?
+		if !args.Dir.Encloses(goMod) {
+			return false
+		}
+
+		// Calculate the relative path
+		rel, err := filepath.Rel(args.Dir.Path(), goMod.Path())
+		if err != nil {
+			return false // "can't happen" (see prior Encloses check)
+		}
+
+		assert(filepath.Base(goMod.Path()) == "go.mod", fmt.Sprintf("invalid go.mod path: want go.mod, got %q", goMod.Path()))
+
+		// Invariant: rel is a relative path without "../" segments and the last
+		// segment is "go.mod"
+		nparts := strings.Count(rel, string(filepath.Separator))
+		return args.MaxDepth < 0 || nparts <= args.MaxDepth
+	}
+
+	// Views may include:
+	//   - go.work views containing one or more modules each;
+	//   - go.mod views containing a single module each;
+	//   - GOPATH and/or ad hoc views containing no modules.
+	//
+	// Retrieving a view via the request path would only work for a
+	// non-recursive query for a go.mod view, and even in that case
+	// [Session.SnapshotOf] doesn't work on directories. Thus we check every
+	// view.
+	var result command.ModulesResult
+	seen := map[protocol.DocumentURI]bool{}
+	for _, v := range h.s.session.Views() {
+		s, release, err := v.Snapshot()
+		if err != nil {
+			return command.ModulesResult{}, err
+		}
+		defer release()
+
+		for _, modFile := range v.ModFiles() {
+			if !keepModule(modFile) {
+				continue
+			}
+
+			// Deduplicate
+			if seen[modFile] {
+				continue
+			}
+			seen[modFile] = true
+
+			fh, err := s.ReadFile(ctx, modFile)
+			if err != nil {
+				return command.ModulesResult{}, err
+			}
+			mod, err := s.ParseMod(ctx, fh)
+			if err != nil {
+				return command.ModulesResult{}, err
+			}
+			if mod.File.Module == nil {
+				continue // syntax contains errors
+			}
+			result.Modules = append(result.Modules, command.Module{
+				Path:    mod.File.Module.Mod.Path,
+				Version: mod.File.Module.Mod.Version,
+				GoMod:   mod.URI,
+			})
+		}
+	}
+	return result, nil
+}
+
+func (h *commandHandler) Packages(ctx context.Context, args command.PackagesArgs) (command.PackagesResult, error) {
+	// Convert file arguments into directories
+	dirs := make([]protocol.DocumentURI, len(args.Files))
+	for i, file := range args.Files {
+		if filepath.Ext(file.Path()) == ".go" {
+			dirs[i] = file.Dir()
+		} else {
+			dirs[i] = file
+		}
+	}
+
+	keepPackage := func(pkg *metadata.Package) bool {
+		for _, file := range pkg.GoFiles {
+			for _, dir := range dirs {
+				if file.Dir() == dir || args.Recursive && dir.Encloses(file) {
+					return true
+				}
+			}
+		}
+		return false
+	}
+
+	result := command.PackagesResult{
+		Module: make(map[string]command.Module),
+	}
+
+	err := h.run(ctx, commandConfig{
+		progress: "Packages",
+	}, func(ctx context.Context, _ commandDeps) error {
+		for _, view := range h.s.session.Views() {
+			snapshot, release, err := view.Snapshot()
+			if err != nil {
+				return err
+			}
+			defer release()
+
+			metas, err := snapshot.WorkspaceMetadata(ctx)
+			if err != nil {
+				return err
+			}
+
+			// Filter out unwanted packages
+			metas = slices.DeleteFunc(metas, func(meta *metadata.Package) bool {
+				return meta.IsIntermediateTestVariant() ||
+					!keepPackage(meta)
+			})
+
+			start := len(result.Packages)
+			for _, meta := range metas {
+				var mod command.Module
+				if meta.Module != nil {
+					mod = command.Module{
+						Path:    meta.Module.Path,
+						Version: meta.Module.Version,
+						GoMod:   protocol.URIFromPath(meta.Module.GoMod),
+					}
+					result.Module[mod.Path] = mod // Overwriting is ok
+				}
+
+				result.Packages = append(result.Packages, command.Package{
+					Path:       string(meta.PkgPath),
+					ForTest:    string(meta.ForTest),
+					ModulePath: mod.Path,
+				})
+			}
+
+			if args.Mode&command.NeedTests == 0 {
+				continue
+			}
+
+			// Make a single request to the index (per snapshot) to minimize the
+			// performance hit
+			var ids []cache.PackageID
+			for _, meta := range metas {
+				ids = append(ids, meta.ID)
+			}
+
+			allTests, err := snapshot.Tests(ctx, ids...)
+			if err != nil {
+				return err
+			}
+
+			for i, tests := range allTests {
+				pkg := &result.Packages[start+i]
+				fileByPath := map[protocol.DocumentURI]*command.TestFile{}
+				for _, test := range tests.All() {
+					test := command.TestCase{
+						Name: test.Name,
+						Loc:  test.Location,
+					}
+
+					file, ok := fileByPath[test.Loc.URI]
+					if !ok {
+						f := command.TestFile{
+							URI: test.Loc.URI,
+						}
+						i := len(pkg.TestFiles)
+						pkg.TestFiles = append(pkg.TestFiles, f)
+						file = &pkg.TestFiles[i]
+						fileByPath[test.Loc.URI] = file
+					}
+					file.Tests = append(file.Tests, test)
+				}
+			}
+		}
+
+		return nil
+	})
+	return result, err
+}
+
 func (h *commandHandler) MaybePromptForTelemetry(ctx context.Context) error {
 	go h.s.maybePromptForTelemetry(ctx, true)
 	return nil
@@ -92,18 +277,8 @@ func (*commandHandler) AddTelemetryCounters(_ context.Context, args command.AddT
 
 // commandConfig configures common command set-up and execution.
 type commandConfig struct {
-	// TODO(adonovan): whether a command is synchronous or
-	// asynchronous is part of the server interface contract, not
-	// a mere implementation detail of the handler.
-	// Export a (command.Command).IsAsync() property so that
-	// clients can tell. (The tricky part is ensuring the handler
-	// remains consistent with the command.Command metadata, as at
-	// the point were we read the 'async' field below, we no
-	// longer know that command.Command.)
-
-	async       bool                 // whether to run the command asynchronously. Async commands can only return errors.
 	requireSave bool                 // whether all files must be saved for the command to work
-	progress    string               // title to use for progress reporting. If empty, no progress will be reported.
+	progress    string               // title to use for progress reporting. If empty, no progress will be reported. Mandatory for async commands.
 	forView     string               // view to resolve to a snapshot; incompatible with forURI
 	forURI      protocol.DocumentURI // URI to resolve to a snapshot. If unset, snapshot will be nil.
 }
@@ -112,9 +287,9 @@ type commandConfig struct {
 // be populated, depending on which configuration is set. See comments in-line
 // for details.
 type commandDeps struct {
-	snapshot *cache.Snapshot    // present if cfg.forURI was set
+	snapshot *cache.Snapshot    // present if cfg.forURI or forView was set
 	fh       file.Handle        // present if cfg.forURI was set
-	work     *progress.WorkDone // present cfg.progress was set
+	work     *progress.WorkDone // present if cfg.progress was set
 }
 
 type commandFunc func(context.Context, commandDeps) error
@@ -194,7 +369,12 @@ func (c *commandHandler) run(ctx context.Context, cfg commandConfig, run command
 		}
 		return err
 	}
-	if cfg.async {
+
+	if enum := command.Command(c.params.Command); enum.IsAsync() {
+		if cfg.progress == "" {
+			log.Fatalf("asynchronous command %q does not enable progress reporting",
+				enum)
+		}
 		go func() {
 			if err := runcmd(); err != nil {
 				showMessage(ctx, c.s.client, protocol.Error, err.Error())
@@ -209,34 +389,25 @@ func (c *commandHandler) ApplyFix(ctx context.Context, args command.ApplyFixArgs
 	var result *protocol.WorkspaceEdit
 	err := c.run(ctx, commandConfig{
 		// Note: no progress here. Applying fixes should be quick.
-		forURI: args.URI,
+		forURI: args.Location.URI,
 	}, func(ctx context.Context, deps commandDeps) error {
-		edits, err := golang.ApplyFix(ctx, args.Fix, deps.snapshot, deps.fh, args.Range)
+		changes, err := golang.ApplyFix(ctx, args.Fix, deps.snapshot, deps.fh, args.Location.Range)
 		if err != nil {
 			return err
 		}
-		changes := []protocol.DocumentChanges{} // must be a slice
-		for _, edit := range edits {
-			edit := edit
-			changes = append(changes, protocol.DocumentChanges{
-				TextDocumentEdit: &edit,
-			})
-		}
-		edit := protocol.WorkspaceEdit{
-			DocumentChanges: changes,
-		}
+		wsedit := protocol.NewWorkspaceEdit(changes...)
 		if args.ResolveEdits {
-			result = &edit
+			result = wsedit
 			return nil
 		}
-		r, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
-			Edit: edit,
+		resp, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
+			Edit: *wsedit,
 		})
 		if err != nil {
 			return err
 		}
-		if !r.Applied {
-			return errors.New(r.FailureReason)
+		if !resp.Applied {
+			return errors.New(resp.FailureReason)
 		}
 		return nil
 	})
@@ -280,7 +451,9 @@ func (c *commandHandler) modifyState(ctx context.Context, source ModificationSou
 	}
 	wg.Add(1)
 	go func() {
-		c.s.diagnoseSnapshot(snapshot, nil, 0)
+		// Diagnosing with the background context ensures new snapshots are fully
+		// diagnosed.
+		c.s.diagnoseSnapshot(snapshot.BackgroundContext(), snapshot, nil, 0)
 		release()
 		wg.Done()
 	}()
@@ -364,8 +537,7 @@ func (c *commandHandler) UpdateGoSum(ctx context.Context, args command.URIArgs)
 
 func (c *commandHandler) Tidy(ctx context.Context, args command.URIArgs) error {
 	return c.run(ctx, commandConfig{
-		requireSave: true,
-		progress:    "Running go mod tidy",
+		progress: "Running go mod tidy",
 	}, func(ctx context.Context, _ commandDeps) error {
 		for _, uri := range args.URIs {
 			fh, snapshot, release, err := c.s.fileOf(ctx, uri)
@@ -386,7 +558,7 @@ func (c *commandHandler) Tidy(ctx context.Context, args command.URIArgs) error {
 
 func (c *commandHandler) Vendor(ctx context.Context, args command.URIArg) error {
 	return c.run(ctx, commandConfig{
-		requireSave: true,
+		requireSave: true, // TODO(adonovan): probably not needed; but needs a test.
 		progress:    "Running go mod vendor",
 		forURI:      args.URI,
 	}, func(ctx context.Context, deps commandDeps) error {
@@ -399,11 +571,16 @@ func (c *commandHandler) Vendor(ctx context.Context, args command.URIArg) error
 		// modules.txt in-place. In that case we could theoretically allow this
 		// command to run concurrently.
 		stderr := new(bytes.Buffer)
-		err := deps.snapshot.RunGoCommandPiped(ctx, cache.Normal|cache.AllowNetwork, &gocommand.Invocation{
+		inv, cleanupInvocation, err := deps.snapshot.GoCommandInvocation(true, &gocommand.Invocation{
 			Verb:       "mod",
 			Args:       []string{"vendor"},
 			WorkingDir: filepath.Dir(args.URI.Path()),
-		}, &bytes.Buffer{}, stderr)
+		})
+		if err != nil {
+			return err
+		}
+		defer cleanupInvocation()
+		err = deps.snapshot.View().GoCommandRunner().RunPiped(ctx, *inv, &bytes.Buffer{}, stderr)
 		if err != nil {
 			return fmt.Errorf("running go mod vendor failed: %v\nstderr:\n%s", err, stderr.String())
 		}
@@ -458,9 +635,8 @@ func (c *commandHandler) RemoveDependency(ctx context.Context, args command.Remo
 			return err
 		}
 		response, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
-			Edit: protocol.WorkspaceEdit{
-				DocumentChanges: documentChanges(deps.fh, edits),
-			},
+			Edit: *protocol.NewWorkspaceEdit(
+				protocol.DocumentChangeEdit(deps.fh, edits)),
 		})
 		if err != nil {
 			return err
@@ -494,6 +670,7 @@ func dropDependency(pm *cache.ParsedModule, modulePath string) ([]protocol.TextE
 	return protocol.EditsFromDiffEdits(pm.Mapper, diff)
 }
 
+// Test is an alias for RunTests (with splayed arguments).
 func (c *commandHandler) Test(ctx context.Context, uri protocol.DocumentURI, tests, benchmarks []string) error {
 	return c.RunTests(ctx, command.RunTestsArgs{
 		URI:        uri,
@@ -502,11 +679,50 @@ func (c *commandHandler) Test(ctx context.Context, uri protocol.DocumentURI, tes
 	})
 }
 
+func (c *commandHandler) Doc(ctx context.Context, args command.DocArgs) (protocol.URI, error) {
+	if args.Location.URI == "" {
+		return "", errors.New("missing location URI")
+	}
+
+	var result protocol.URI
+	err := c.run(ctx, commandConfig{
+		progress: "", // the operation should be fast
+		forURI:   args.Location.URI,
+	}, func(ctx context.Context, deps commandDeps) error {
+		pkg, pgf, err := golang.NarrowestPackageForFile(ctx, deps.snapshot, args.Location.URI)
+		if err != nil {
+			return err
+		}
+		start, end, err := pgf.RangePos(args.Location.Range)
+		if err != nil {
+			return err
+		}
+
+		// Start web server.
+		web, err := c.s.getWeb()
+		if err != nil {
+			return err
+		}
+
+		// Compute package path and optional symbol fragment
+		// (e.g. "#Buffer.Len") from the the selection.
+		pkgpath, fragment, _ := golang.DocFragment(pkg, pgf, start, end)
+
+		// Direct the client to open the /pkg page.
+		result = web.PkgURL(deps.snapshot.View().ID(), pkgpath, fragment)
+		if args.ShowDocument {
+			openClientBrowser(ctx, c.s.client, result)
+		}
+
+		return nil
+	})
+	return result, err
+}
+
 func (c *commandHandler) RunTests(ctx context.Context, args command.RunTestsArgs) error {
 	return c.run(ctx, commandConfig{
-		async:       true,
-		progress:    "Running go test",
-		requireSave: true,
+		progress:    "Running go test", // (asynchronous)
+		requireSave: true,              // go test honors overlays, but tests themselves cannot
 		forURI:      args.URI,
 	}, func(ctx context.Context, deps commandDeps) error {
 		return c.runTests(ctx, deps.snapshot, deps.work, args.URI, args.Tests, args.Benchmarks)
@@ -529,12 +745,16 @@ func (c *commandHandler) runTests(ctx context.Context, snapshot *cache.Snapshot,
 	// Run `go test -run Func` on each test.
 	var failedTests int
 	for _, funcName := range tests {
-		inv := &gocommand.Invocation{
+		inv, cleanupInvocation, err := snapshot.GoCommandInvocation(false, &gocommand.Invocation{
 			Verb:       "test",
 			Args:       []string{pkgPath, "-v", "-count=1", fmt.Sprintf("-run=^%s$", regexp.QuoteMeta(funcName))},
 			WorkingDir: filepath.Dir(uri.Path()),
+		})
+		if err != nil {
+			return err
 		}
-		if err := snapshot.RunGoCommandPiped(ctx, cache.Normal, inv, out, out); err != nil {
+		defer cleanupInvocation()
+		if err := snapshot.View().GoCommandRunner().RunPiped(ctx, *inv, out, out); err != nil {
 			if errors.Is(err, context.Canceled) {
 				return err
 			}
@@ -545,12 +765,16 @@ func (c *commandHandler) runTests(ctx context.Context, snapshot *cache.Snapshot,
 	// Run `go test -run=^$ -bench Func` on each test.
 	var failedBenchmarks int
 	for _, funcName := range benchmarks {
-		inv := &gocommand.Invocation{
+		inv, cleanupInvocation, err := snapshot.GoCommandInvocation(false, &gocommand.Invocation{
 			Verb:       "test",
 			Args:       []string{pkgPath, "-v", "-run=^$", fmt.Sprintf("-bench=^%s$", regexp.QuoteMeta(funcName))},
 			WorkingDir: filepath.Dir(uri.Path()),
+		})
+		if err != nil {
+			return err
 		}
-		if err := snapshot.RunGoCommandPiped(ctx, cache.Normal, inv, out, out); err != nil {
+		defer cleanupInvocation()
+		if err := snapshot.View().GoCommandRunner().RunPiped(ctx, *inv, out, out); err != nil {
 			if errors.Is(err, context.Canceled) {
 				return err
 			}
@@ -594,7 +818,7 @@ func (c *commandHandler) Generate(ctx context.Context, args command.GenerateArgs
 		title = "Running go generate ./..."
 	}
 	return c.run(ctx, commandConfig{
-		requireSave: true,
+		requireSave: true, // commands executed by go generate cannot honor overlays
 		progress:    title,
 		forURI:      args.Dir,
 	}, func(ctx context.Context, deps commandDeps) error {
@@ -604,13 +828,17 @@ func (c *commandHandler) Generate(ctx context.Context, args command.GenerateArgs
 		if args.Recursive {
 			pattern = "./..."
 		}
-		inv := &gocommand.Invocation{
+		inv, cleanupInvocation, err := deps.snapshot.GoCommandInvocation(true, &gocommand.Invocation{
 			Verb:       "generate",
 			Args:       []string{"-x", pattern},
 			WorkingDir: args.Dir.Path(),
+		})
+		if err != nil {
+			return err
 		}
+		defer cleanupInvocation()
 		stderr := io.MultiWriter(er, progress.NewWorkDoneWriter(ctx, deps.work))
-		if err := deps.snapshot.RunGoCommandPiped(ctx, cache.AllowNetwork, inv, er, stderr); err != nil {
+		if err := deps.snapshot.View().GoCommandRunner().RunPiped(ctx, *inv, er, stderr); err != nil {
 			return err
 		}
 		return nil
@@ -622,17 +850,33 @@ func (c *commandHandler) GoGetPackage(ctx context.Context, args command.GoGetPac
 		forURI:   args.URI,
 		progress: "Running go get",
 	}, func(ctx context.Context, deps commandDeps) error {
-		// Run on a throwaway go.mod, otherwise it'll write to the real one.
-		stdout, err := deps.snapshot.RunGoCommandDirect(ctx, cache.WriteTemporaryModFile|cache.AllowNetwork, &gocommand.Invocation{
+		snapshot := deps.snapshot
+		modURI := snapshot.GoModForFile(args.URI)
+		if modURI == "" {
+			return fmt.Errorf("no go.mod file found for %s", args.URI)
+		}
+		tempDir, cleanupModDir, err := cache.TempModDir(ctx, snapshot, modURI)
+		if err != nil {
+			return fmt.Errorf("creating a temp go.mod: %v", err)
+		}
+		defer cleanupModDir()
+
+		inv, cleanupInvocation, err := snapshot.GoCommandInvocation(true, &gocommand.Invocation{
 			Verb:       "list",
-			Args:       []string{"-f", "{{.Module.Path}}@{{.Module.Version}}", args.Pkg},
-			WorkingDir: filepath.Dir(args.URI.Path()),
+			Args:       []string{"-f", "{{.Module.Path}}@{{.Module.Version}}", "-mod=mod", "-modfile=" + filepath.Join(tempDir, "go.mod"), args.Pkg},
+			Env:        []string{"GOWORK=off"},
+			WorkingDir: modURI.Dir().Path(),
 		})
 		if err != nil {
 			return err
 		}
+		defer cleanupInvocation()
+		stdout, err := snapshot.View().GoCommandRunner().Run(ctx, *inv)
+		if err != nil {
+			return err
+		}
 		ver := strings.TrimSpace(stdout.String())
-		return c.s.runGoModUpdateCommands(ctx, deps.snapshot, args.URI, func(invoke func(...string) (*bytes.Buffer, error)) error {
+		return c.s.runGoModUpdateCommands(ctx, snapshot, args.URI, func(invoke func(...string) (*bytes.Buffer, error)) error {
 			if args.AddRequire {
 				if err := addModuleRequire(invoke, []string{ver}); err != nil {
 					return err
@@ -645,43 +889,57 @@ func (c *commandHandler) GoGetPackage(ctx context.Context, args command.GoGetPac
 }
 
 func (s *server) runGoModUpdateCommands(ctx context.Context, snapshot *cache.Snapshot, uri protocol.DocumentURI, run func(invoke func(...string) (*bytes.Buffer, error)) error) error {
-	newModBytes, newSumBytes, err := snapshot.RunGoModUpdateCommands(ctx, filepath.Dir(uri.Path()), run)
-	if err != nil {
-		return err
-	}
+	// TODO(rfindley): can/should this use findRootPattern?
 	modURI := snapshot.GoModForFile(uri)
+	if modURI == "" {
+		return fmt.Errorf("no go.mod file found for %s", uri.Path())
+	}
+	newModBytes, newSumBytes, err := snapshot.RunGoModUpdateCommands(ctx, modURI, run)
+	if err != nil {
+		return err
+	}
 	sumURI := protocol.URIFromPath(strings.TrimSuffix(modURI.Path(), ".mod") + ".sum")
-	modEdits, err := collectFileEdits(ctx, snapshot, modURI, newModBytes)
+
+	modChange, err := computeEditChange(ctx, snapshot, modURI, newModBytes)
 	if err != nil {
 		return err
 	}
-	sumEdits, err := collectFileEdits(ctx, snapshot, sumURI, newSumBytes)
+	sumChange, err := computeEditChange(ctx, snapshot, sumURI, newSumBytes)
 	if err != nil {
 		return err
 	}
-	return applyFileEdits(ctx, s.client, append(sumEdits, modEdits...))
+
+	var changes []protocol.DocumentChange
+	if modChange.Valid() {
+		changes = append(changes, modChange)
+	}
+	if sumChange.Valid() {
+		changes = append(changes, sumChange)
+	}
+	return applyChanges(ctx, s.client, changes)
 }
 
-// collectFileEdits collects any file edits required to transform the snapshot
-// file specified by uri to the provided new content.
+// computeEditChange computes the edit change required to transform the
+// snapshot file specified by uri to the provided new content.
+// Beware: returns a DocumentChange that is !Valid() if none were necessary.
 //
-// If the file is not open, collectFileEdits simply writes the new content to
+// If the file is not open, computeEditChange simply writes the new content to
 // disk.
 //
 // TODO(rfindley): fix this API asymmetry. It should be up to the caller to
 // write the file or apply the edits.
-func collectFileEdits(ctx context.Context, snapshot *cache.Snapshot, uri protocol.DocumentURI, newContent []byte) ([]protocol.TextDocumentEdit, error) {
+func computeEditChange(ctx context.Context, snapshot *cache.Snapshot, uri protocol.DocumentURI, newContent []byte) (protocol.DocumentChange, error) {
 	fh, err := snapshot.ReadFile(ctx, uri)
 	if err != nil {
-		return nil, err
+		return protocol.DocumentChange{}, err
 	}
 	oldContent, err := fh.Content()
 	if err != nil && !os.IsNotExist(err) {
-		return nil, err
+		return protocol.DocumentChange{}, err
 	}
 
 	if bytes.Equal(oldContent, newContent) {
-		return nil, nil
+		return protocol.DocumentChange{}, nil // note: result is !Valid()
 	}
 
 	// Sending a workspace edit to a closed file causes VS Code to open the
@@ -689,34 +947,24 @@ func collectFileEdits(ctx context.Context, snapshot *cache.Snapshot, uri protoco
 	// especially to go.sum, which should be mostly invisible to the user.
 	if !snapshot.IsOpen(uri) {
 		err := os.WriteFile(uri.Path(), newContent, 0666)
-		return nil, err
+		return protocol.DocumentChange{}, err
 	}
 
 	m := protocol.NewMapper(fh.URI(), oldContent)
 	diff := diff.Bytes(oldContent, newContent)
-	edits, err := protocol.EditsFromDiffEdits(m, diff)
+	textedits, err := protocol.EditsFromDiffEdits(m, diff)
 	if err != nil {
-		return nil, err
+		return protocol.DocumentChange{}, err
 	}
-	return []protocol.TextDocumentEdit{{
-		TextDocument: protocol.OptionalVersionedTextDocumentIdentifier{
-			Version: fh.Version(),
-			TextDocumentIdentifier: protocol.TextDocumentIdentifier{
-				URI: uri,
-			},
-		},
-		Edits: protocol.AsAnnotatedTextEdits(edits),
-	}}, nil
+	return protocol.DocumentChangeEdit(fh, textedits), nil
 }
 
-func applyFileEdits(ctx context.Context, cli protocol.Client, edits []protocol.TextDocumentEdit) error {
-	if len(edits) == 0 {
+func applyChanges(ctx context.Context, cli protocol.Client, changes []protocol.DocumentChange) error {
+	if len(changes) == 0 {
 		return nil
 	}
 	response, err := cli.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
-		Edit: protocol.WorkspaceEdit{
-			DocumentChanges: protocol.TextDocumentEditsToDocumentChanges(edits),
-		},
+		Edit: *protocol.NewWorkspaceEdit(changes...),
 	})
 	if err != nil {
 		return err
@@ -748,14 +996,20 @@ func addModuleRequire(invoke func(...string) (*bytes.Buffer, error), args []stri
 
 // TODO(rfindley): inline.
 func (s *server) getUpgrades(ctx context.Context, snapshot *cache.Snapshot, uri protocol.DocumentURI, modules []string) (map[string]string, error) {
-	stdout, err := snapshot.RunGoCommandDirect(ctx, cache.Normal|cache.AllowNetwork, &gocommand.Invocation{
-		Verb:       "list",
-		Args:       append([]string{"-m", "-u", "-json"}, modules...),
+	inv, cleanup, err := snapshot.GoCommandInvocation(true, &gocommand.Invocation{
+		Verb: "list",
+		// -mod=readonly is necessary when vendor is present (golang/go#66055)
+		Args:       append([]string{"-mod=readonly", "-m", "-u", "-json"}, modules...),
 		WorkingDir: filepath.Dir(uri.Path()),
 	})
 	if err != nil {
 		return nil, err
 	}
+	defer cleanup()
+	stdout, err := snapshot.View().GoCommandRunner().Run(ctx, *inv)
+	if err != nil {
+		return nil, err
+	}
 
 	upgrades := map[string]string{}
 	for dec := json.NewDecoder(stdout); dec.More(); {
@@ -777,9 +1031,8 @@ func (c *commandHandler) GCDetails(ctx context.Context, uri protocol.DocumentURI
 
 func (c *commandHandler) ToggleGCDetails(ctx context.Context, args command.URIArg) error {
 	return c.run(ctx, commandConfig{
-		requireSave: true,
-		progress:    "Toggling GC Details",
-		forURI:      args.URI,
+		progress: "Toggling GC Details",
+		forURI:   args.URI,
 	}, func(ctx context.Context, deps commandDeps) error {
 		return c.modifyState(ctx, FromToggleGCDetails, func() (*cache.Snapshot, func(), error) {
 			meta, err := golang.NarrowestMetadataForFile(ctx, deps.snapshot, deps.fh.URI())
@@ -865,13 +1118,36 @@ func (c *commandHandler) AddImport(ctx context.Context, args command.AddImportAr
 		if err != nil {
 			return fmt.Errorf("could not add import: %v", err)
 		}
-		if _, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
-			Edit: protocol.WorkspaceEdit{
-				DocumentChanges: documentChanges(deps.fh, edits),
-			},
-		}); err != nil {
+		r, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
+			Edit: *protocol.NewWorkspaceEdit(
+				protocol.DocumentChangeEdit(deps.fh, edits)),
+		})
+		if err != nil {
 			return fmt.Errorf("could not apply import edits: %v", err)
 		}
+		if !r.Applied {
+			return fmt.Errorf("failed to apply edits: %v", r.FailureReason)
+		}
+		return nil
+	})
+}
+
+func (c *commandHandler) ExtractToNewFile(ctx context.Context, args protocol.Location) error {
+	return c.run(ctx, commandConfig{
+		progress: "Extract to a new file",
+		forURI:   args.URI,
+	}, func(ctx context.Context, deps commandDeps) error {
+		edit, err := golang.ExtractToNewFile(ctx, deps.snapshot, deps.fh, args.Range)
+		if err != nil {
+			return err
+		}
+		resp, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{Edit: *edit})
+		if err != nil {
+			return fmt.Errorf("could not apply edits: %v", err)
+		}
+		if !resp.Applied {
+			return fmt.Errorf("edits not applied: %s", resp.FailureReason)
+		}
 		return nil
 	})
 }
@@ -969,9 +1245,8 @@ func (c *commandHandler) RunGovulncheck(ctx context.Context, args command.Vulnch
 	// synchronize the start of the run and return the token.
 	tokenChan := make(chan protocol.ProgressToken, 1)
 	err := c.run(ctx, commandConfig{
-		async:       true, // need to be async to be cancellable
-		progress:    "govulncheck",
-		requireSave: true,
+		progress:    "govulncheck", // (asynchronous)
+		requireSave: true,          // govulncheck cannot honor overlays
 		forURI:      args.URI,
 	}, func(ctx context.Context, deps commandDeps) error {
 		tokenChan <- deps.work.Token()
@@ -992,7 +1267,10 @@ func (c *commandHandler) RunGovulncheck(ctx context.Context, args command.Vulnch
 			return err
 		}
 		defer release()
-		c.s.diagnoseSnapshot(snapshot, nil, 0)
+
+		// Diagnosing with the background context ensures new snapshots are fully
+		// diagnosed.
+		c.s.diagnoseSnapshot(snapshot.BackgroundContext(), snapshot, nil, 0)
 
 		affecting := make(map[string]bool, len(result.Entries))
 		for _, finding := range result.Findings {
@@ -1224,6 +1502,9 @@ func openClientBrowser(ctx context.Context, cli protocol.Client, url protocol.UR
 
 // openClientEditor causes the LSP client to open the specified document
 // and select the indicated range.
+//
+// Note that VS Code 1.87.2 doesn't currently raise the window; this is
+// https://github.com/microsoft/vscode/issues/207634
 func openClientEditor(ctx context.Context, cli protocol.Client, loc protocol.Location) {
 	showDocumentImpl(ctx, cli, protocol.URI(loc.URI), &loc.Range)
 }
@@ -1266,24 +1547,21 @@ func (c *commandHandler) ChangeSignature(ctx context.Context, args command.Chang
 		forURI: args.RemoveParameter.URI,
 	}, func(ctx context.Context, deps commandDeps) error {
 		// For now, gopls only supports removing unused parameters.
-		changes, err := golang.RemoveUnusedParameter(ctx, deps.fh, args.RemoveParameter.Range, deps.snapshot)
+		docedits, err := golang.RemoveUnusedParameter(ctx, deps.fh, args.RemoveParameter.Range, deps.snapshot)
 		if err != nil {
 			return err
 		}
-		edit := protocol.WorkspaceEdit{
-			DocumentChanges: changes,
-		}
+		wsedit := protocol.NewWorkspaceEdit(docedits...)
 		if args.ResolveEdits {
-			result = &edit
+			result = wsedit
 			return nil
 		}
 		r, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
-			Edit: edit,
+			Edit: *wsedit,
 		})
 		if !r.Applied {
 			return fmt.Errorf("failed to apply edits: %v", r.FailureReason)
 		}
-
 		return nil
 	})
 	return result, err
@@ -1321,7 +1599,11 @@ func (c *commandHandler) DiagnoseFiles(ctx context.Context, args command.Diagnos
 			wg.Add(1)
 			go func() {
 				defer wg.Done()
-				c.s.diagnoseSnapshot(snapshot, nil, 0)
+
+				// Use the operation context for diagnosis, rather than
+				// snapshot.BackgroundContext, because this operation does not create
+				// new snapshots (so they should also be diagnosed by other means).
+				c.s.diagnoseSnapshot(ctx, snapshot, nil, 0)
 			}()
 		}
 		wg.Wait()
@@ -1334,6 +1616,7 @@ func (c *commandHandler) Views(ctx context.Context) ([]command.View, error) {
 	var summaries []command.View
 	for _, view := range c.s.session.Views() {
 		summaries = append(summaries, command.View{
+			ID:         view.ID(),
 			Type:       view.Type().String(),
 			Root:       view.Root(),
 			Folder:     view.Folder().Dir,
@@ -1342,3 +1625,35 @@ func (c *commandHandler) Views(ctx context.Context) ([]command.View, error) {
 	}
 	return summaries, nil
 }
+
+func (c *commandHandler) FreeSymbols(ctx context.Context, viewID string, loc protocol.Location) error {
+	web, err := c.s.getWeb()
+	if err != nil {
+		return err
+	}
+	url := web.freesymbolsURL(viewID, loc)
+	openClientBrowser(ctx, c.s.client, url)
+	return nil
+}
+
+func (c *commandHandler) Assembly(ctx context.Context, viewID, packageID, symbol string) error {
+	web, err := c.s.getWeb()
+	if err != nil {
+		return err
+	}
+	url := web.assemblyURL(viewID, packageID, symbol)
+	openClientBrowser(ctx, c.s.client, url)
+	return nil
+}
+
+func (c *commandHandler) ClientOpenURL(ctx context.Context, url string) error {
+	openClientBrowser(ctx, c.s.client, url)
+	return nil
+}
+
+func (c *commandHandler) ScanImports(ctx context.Context) error {
+	for _, v := range c.s.session.Views() {
+		v.ScanImports()
+	}
+	return nil
+}
